/**
 ****************************************************************************************
 *
 * @file led_switch_task.c
 *
 * @brief LED Switch Service Task implementation.
 *
 * Copyright (C) RivieraWaves 2009-2016
 *
 *
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @addtogroup LED_SWITCH_TASK
 * @{
 ****************************************************************************************
 */


/*
 * INCLUDE FILES
 ****************************************************************************************
 */
#include "rwip_config.h"

#if (BLE_LED_SWITCH_SERVER)
#include "gap.h"
#include "gattc_task.h"
#include "app.h"
#include "led_switch.h"
#include "led_switch_task.h"
#include "prf_utils.h"
#include "ke_mem.h"
#include "uart.h"

extern bool uart_tp_mode_flag;
/*
 * FUNCTION DEFINITIONS
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @brief Handles reception of the @ref LED_SWITCH_SET_VALUE_REQ message.
 *
 * @param[in] msgid Id of the message received (probably unused).
 * @param[in] param Pointer to the parameters of the message.
 * @param[in] dest_id ID of the receiving task instance (probably unused).
 * @param[in] src_id ID of the sending task instance.
 * @return If the message was consumed or not.
 ****************************************************************************************
 */
static int led_switch_set_value_req_handler(ke_msg_id_t const msgid,
                                      struct led_switch_set_value_req const *param,
                                      ke_task_id_t const dest_id,
                                      ke_task_id_t const src_id)
{
    // Request status
    uint8_t status;
    // Characteristic Declaration attribute handle
    uint16_t handle;

    struct led_switch_env_tag* led_switch_env = PRF_ENV_GET(LED_SWITCH, led_switch);
    struct led_switch_set_value_rsp* rsp;
    // Check Characteristic Code
    if (param->value < LED_SWITCH_CHAR_MAX)
    {
        // Get Characteristic Declaration attribute handle
        handle = led_switch_value_to_handle(led_switch_env, param->value);

        // Check if the Characteristic exists in the database
        if (handle != ATT_INVALID_HDL)
        {
            // Check the value length
            status = led_switch_check_val_len(param->value, param->length);

            if (status == GAP_ERR_NO_ERROR)
            {
                // Check value in already present in service
                struct led_switch_val_elmt *val = (struct led_switch_val_elmt *) co_list_pick(&(led_switch_env->values));
                // loop until value found
                while(val != NULL)
                {
                    // if value already present, remove old one
                    if(val->value == param->value)
                    {
                        co_list_extract(&(led_switch_env->values), &(val->hdr), 0);
                        ke_free(val);
                        break;
                    }
                    val = (struct led_switch_val_elmt *)val->hdr.next;
                }

                // allocate value data
                val = (struct led_switch_val_elmt *) ke_malloc(sizeof(struct led_switch_val_elmt) + param->length, KE_MEM_ATT_DB);
                val->value = param->value;
                val->length = param->length;
                memcpy(val->data, param->data, param->length);
                // insert value into the list
                co_list_push_back(&(led_switch_env->values), &(val->hdr));
            }
        }
        else
        {
            status = PRF_ERR_INEXISTENT_HDL;
        }
    }
    else
    {
        status = PRF_ERR_INVALID_PARAM;
    }

    // send response to application
    rsp = KE_MSG_ALLOC(LED_SWITCH_SET_VALUE_RSP, src_id, dest_id, led_switch_set_value_rsp);
    rsp->value = param->value;
    rsp->status = status;
    ke_msg_send(rsp);


    return (KE_MSG_CONSUMED);
}

static int gattc_write_req_ind_handler(ke_msg_id_t const msgid, struct gattc_write_req_ind const *param,
                                      ke_task_id_t const dest_id, ke_task_id_t const src_id)
{
    struct gattc_write_cfm * cfm;
    uint8_t status = GAP_ERR_NO_ERROR;
    //Send AT command
    if(uart_tp_mode_flag == false)
    {
      UART_SEND_DATA(AT_CMD_IO_SET, param->value[0]);
    }
  
    cfm = KE_MSG_ALLOC(GATTC_WRITE_CFM, src_id, dest_id, gattc_write_cfm);
    cfm->handle = param->handle;
    cfm->status = status;
    ke_msg_send(cfm);
  
    return (KE_MSG_CONSUMED);
}

/**
 ****************************************************************************************
 * @brief Handles reception of the read request from peer device
 *
 * @param[in] msgid Id of the message received (probably unused).
 * @param[in] param Pointer to the parameters of the message.
 * @param[in] dest_id ID of the receiving task instance (probably unused).
 * @param[in] src_id ID of the sending task instance.
 * @return If the message was consumed or not.
 ****************************************************************************************
 */
static int gattc_read_req_ind_handler(ke_msg_id_t const msgid,
                                      struct gattc_read_req_ind const *param,
                                      ke_task_id_t const dest_id,
                                      ke_task_id_t const src_id)
{
    int msg_status = KE_MSG_CONSUMED;
    ke_state_t state = ke_state_get(dest_id);

    if(state == LED_SWITCH_IDLE)
    {
        struct led_switch_env_tag* led_switch_env = PRF_ENV_GET(LED_SWITCH, led_switch);
        // retrieve value attribute
        uint8_t value = led_switch_handle_to_value(led_switch_env, param->handle);

        // Check Characteristic Code
        if (value < LED_SWITCH_CHAR_MAX)
        {
            // Check value in already present in service
            struct led_switch_val_elmt *val = (struct led_switch_val_elmt *) co_list_pick(&(led_switch_env->values));
            // loop until value found
            while(val != NULL)
            {
                // value is present in service
                if(val->value == value)
                {
                    break;
                }
                val = (struct led_switch_val_elmt *)val->hdr.next;
            }

            if(val != NULL)
            {
                // Send value to peer device.
                struct gattc_read_cfm* cfm = KE_MSG_ALLOC_DYN(GATTC_READ_CFM, src_id, dest_id, gattc_read_cfm, val->length);
                cfm->handle = param->handle;
                cfm->status = ATT_ERR_NO_ERROR;
                cfm->length = val->length;
                memcpy(cfm->value, val->data, val->length);
                ke_msg_send(cfm);
            }
            else
            {
                // request value to application
                led_switch_env->req_val    = value;
                led_switch_env->req_conidx = KE_IDX_GET(src_id);

                struct led_switch_value_req_ind* req_ind = KE_MSG_ALLOC(LED_SWITCH_VALUE_REQ_IND,
                        prf_dst_task_get(&(led_switch_env->prf_env), KE_IDX_GET(src_id)),
                        dest_id, led_switch_value_req_ind);
                req_ind->value = value;
                ke_msg_send(req_ind);

                // Put Service in a busy state
                ke_state_set(dest_id, LED_SWITCH_BUSY);
            }
        }
        else
        {
            // application error, value cannot be retrieved
            struct gattc_read_cfm* cfm = KE_MSG_ALLOC(GATTC_READ_CFM, src_id, dest_id, gattc_read_cfm);
            cfm->handle = param->handle;
            cfm->status = ATT_ERR_APP_ERROR;
            ke_msg_send(cfm);
        }
    }
    // postpone request if profile is in a busy state - required for multipoint
    else if(state == LED_SWITCH_BUSY)
    {
        msg_status = KE_MSG_SAVED;
    }

    return (msg_status);
}


/**
 ****************************************************************************************
 * @brief Handles reception of the value confirmation from application
 *
 * @param[in] msgid Id of the message received (probably unused).
 * @param[in] param Pointer to the parameters of the message.
 * @param[in] dest_id ID of the receiving task instance (probably unused).
 * @param[in] src_id ID of the sending task instance.
 * @return If the message was consumed or not.
 ****************************************************************************************
 */
static int led_switch_value_cfm_handler(ke_msg_id_t const msgid,
                                      struct led_switch_value_cfm const *param,
                                      ke_task_id_t const dest_id,
                                      ke_task_id_t const src_id)
{
    ke_state_t state = ke_state_get(dest_id);

    if(state == LED_SWITCH_BUSY)
    {
        struct led_switch_env_tag* led_switch_env = PRF_ENV_GET(LED_SWITCH, led_switch);
        // retrieve value attribute
        uint16_t handle = led_switch_value_to_handle(led_switch_env, led_switch_env->req_val);

        // chack if application provide correct value
        if (led_switch_env->req_val == param->value)
        {
            // Send value to peer device.
            struct gattc_read_cfm* cfm = KE_MSG_ALLOC_DYN(GATTC_READ_CFM, KE_BUILD_ID(TASK_GATTC, led_switch_env->req_conidx), dest_id, gattc_read_cfm, param->length);
            cfm->handle = handle;
            cfm->status = ATT_ERR_NO_ERROR;
            cfm->length = param->length;
            memcpy(cfm->value, param->data, param->length);
            ke_msg_send(cfm);
        }
        else
        {
            // application error, value provided by application is not the expected one
            struct gattc_read_cfm* cfm = KE_MSG_ALLOC(GATTC_READ_CFM, KE_BUILD_ID(TASK_GATTC, led_switch_env->req_conidx), dest_id, gattc_read_cfm);
            cfm->handle = handle;
            cfm->status = ATT_ERR_APP_ERROR;
            ke_msg_send(cfm);
        }

        // return to idle state
        ke_state_set(dest_id, LED_SWITCH_IDLE);
    }
    // else ignore request if not in busy state

    return (KE_MSG_CONSUMED);
}

/*
 * GLOBAL VARIABLE DEFINITIONS
 ****************************************************************************************
 */

/// Default State handlers definition
const struct ke_msg_handler led_switch_default_state[] =
{
    {LED_SWITCH_SET_VALUE_REQ,      (ke_msg_func_t)led_switch_set_value_req_handler},
    {GATTC_WRITE_REQ_IND,           (ke_msg_func_t)gattc_write_req_ind_handler},
    {GATTC_READ_REQ_IND,            (ke_msg_func_t)gattc_read_req_ind_handler},
    {LED_SWITCH_VALUE_CFM,          (ke_msg_func_t)led_switch_value_cfm_handler},
};

///Specifies the message handlers that are common to all states.
const struct ke_state_handler led_switch_default_handler = KE_STATE_HANDLER(led_switch_default_state);

#endif //BLE_LED_SWITCH_SERVER

/// @} CUSTOMTASK
